def reverse_bits(bitmask: int, num_bits: int) -> int:
    """Reverse the bits in a bitmask.

    For example bitmask=0b1 reverses to:

    * 0b1 (no change) if num_bits = 1
    * 0b10 if num_bits = 2
    * 0b10000 if num_bits = 5

    Parameters
    ----------
    bitmask : int
        The integer bitmask to reverse
    num_bits : int
        The maximum number of bits in bitmask

    Raises ValueError
        If num_bits < 1, bitmask < 1, or bitmask > (1 << num_bits) - 1
    """
    if num_bits < 1:
        raise ValueError(f"{num_bits} < 1")
    max_value = (1 << num_bits) - 1
    if bitmask == 0:
        return 0
    if bitmask < 0:
        raise ValueError(f"{bitmask=} < 0")
    if bitmask > max_value:
        raise ValueError(f"{bitmask=} > {max_value}")

    input_binstr = f"{bitmask:0{num_bits}b}"
    return int(input_binstr[::-1], base=2)


def bytes_from_shaft_word(shaft_word: int, num_bytes: int) -> bytes:
    """Convert a shaft word to the bytes format needed by Toika

    Parameters
    ----------
    shaft_word : int
        Shaft word as a bit mask (bit N high if shaft N should rise)
    num_bytes: int
        The number of bytes in the shaft word = num_shafts / 8

    Returned bytes = the shaft word as a big-endian bytes,
    but with the bits in each byte reversed.
    This is implemented by reversing all bytes and then
    writing out the resulting integer as a little-endian bytes.
    """
    reversed_shaft_word = reverse_bits(bitmask=shaft_word, num_bits=num_bytes * 8)
    return reversed_shaft_word.to_bytes(length=num_bytes, byteorder="little")


def shaft_word_from_bytes(data_bytes: bytes, num_bytes: int) -> int:
    """Convert a bytes in the format needed by Toika to a shaft word.

    Parameters
    ----------
    data_bytes : bytes
        Shaft bytes string in Toika format.
    num_bytes: int
        The number of bytes in the shaft word = num_shafts / 8

    See bytes_from_shaft_word for the data format.
    """
    bitmask = int.from_bytes(data_bytes, byteorder="little")
    return reverse_bits(bitmask=bitmask, num_bits=num_bytes * 8)



from base_loom_server.testutils import BaseTestLoomServer

from toika_loom_server.main import app

# No need to reduce MockLoom.motion_duration to speed up tests
# because the Toika loom does not report motion state.


class TestLoomServer(BaseTestLoomServer):
    app = app
    extra_args = ()  # ("--direction-control", "loom")



import random

import pytest

from toika_loom_server.utils import (
    bytes_from_shaft_word,
    reverse_bits,
    shaft_word_from_bytes,
)

rnd = random.Random(25)


def get_random_word(num_bits: int) -> int:
    return rnd.randrange(1, 1 << num_bits)


async def test_reverse_bits() -> None:
    assert 0 == reverse_bits(0, num_bits=1)
    assert 1 == reverse_bits(1, num_bits=1)
    for num_bits in (7, 8, 9, 15, 16, 17, 31, 32):
        assert 0 == reverse_bits(0, num_bits=num_bits)
        assert 1 << (num_bits - 1) == reverse_bits(1, num_bits=num_bits)
        assert 1 == reverse_bits(1 << (num_bits - 1), num_bits=num_bits)
        for _ in range(100):
            value = get_random_word(num_bits=num_bits)
            reversed_value = reverse_bits(value, num_bits=num_bits)
            valuestr = f"{value:0{num_bits}b}"
            reversed_valuestr = f"{reversed_value:0{num_bits}b}"
            assert valuestr[::-1] == reversed_valuestr
        with pytest.raises(ValueError):
            reverse_bits(0, num_bits=0)
        with pytest.raises(ValueError):
            reverse_bits(1 << num_bits, num_bits=num_bits)
        with pytest.raises(ValueError):
            reverse_bits(-1, num_bits=num_bits)


async def test_bytes_from_shaft_word() -> None:
    for num_bytes in (1, 2, 3, 4):
        num_bits = num_bytes * 8
        assert int(0).to_bytes(length=num_bytes) == bytes_from_shaft_word(
            0, num_bytes=num_bytes
        )
        # bytes_from_shaft_word has the effect of reversing
        # the bits within each byte, so try doing this manually
        # one bit at a time
        for bitnum in range(8):
            for offset_bytes in range(num_bytes):
                value_no_offset = 1 << bitnum
                value = value_no_offset << (offset_bytes * 8)
                reversed_value_no_offset = reverse_bits(value_no_offset, num_bits=8)
                reversed_value = reversed_value_no_offset << (offset_bytes * 8)
                assert value.to_bytes(length=num_bytes) == bytes_from_shaft_word(
                    reversed_value, num_bytes=num_bytes
                )
                assert reversed_value.to_bytes(
                    length=num_bytes
                ) == bytes_from_shaft_word(value, num_bytes=num_bytes)
                assert (
                    shaft_word_from_bytes(
                        bytes_from_shaft_word(value, num_bytes=num_bytes),
                        num_bytes=num_bytes,
                    )
                    == value
                )
                assert (
                    shaft_word_from_bytes(
                        bytes_from_shaft_word(reversed_value, num_bytes=num_bytes),
                        num_bytes=num_bytes,
                    )
                    == reversed_value
                )
        for _ in range(100):
            shaft_word = get_random_word(num_bits=num_bits)
            shaft_bytes = bytes_from_shaft_word(shaft_word, num_bytes=num_bytes)

            # Compute the expected value by reversing the bits within each byte
            shaft_word_as_bytes = shaft_word.to_bytes(length=num_bytes, byteorder="big")
            expected_bytes = bytes(
                bytearray(reverse_bits(elt, num_bits=8) for elt in shaft_word_as_bytes)
            )
            assert expected_bytes == shaft_bytes

            assert shaft_word == shaft_word_from_bytes(shaft_bytes, num_bytes=num_bytes)



__all__ = ["MockLoom"]

from base_loom_server.base_mock_loom import BaseMockLoom

from toika_loom_server.utils import shaft_word_from_bytes


class MockLoom(BaseMockLoom):
    """Simulate a Toika ES dobby loom.

    Parameters
    ----------
    verbose : bool
        If True, log diagnostic information.

    The user controls this loom by:

    * Call command_reader.create_writer() to create a command writer.
    * Call reply_writer.create_reader() to create a reply reader.
    * Read replies from the reply reader.
    * Write commands to the command writer.
    """

    terminator = b""

    def __init__(self, num_shafts: int, verbose: bool = True) -> None:
        if num_shafts % 8 != 0:
            raise ValueError(f"num_shafts={num_shafts} must be a multiple of 8")
        super().__init__(num_shafts=num_shafts, verbose=verbose)

    async def basic_read(self) -> bytes:
        """Read one command to the loom.

        Perform no error checking, except that self.reader exists.
        """
        assert self.reader is not None  # make mypy happy
        return await self.reader.readexactly(self.num_shafts // 8)

    async def handle_read_bytes(self, read_bytes: bytes) -> None:
        """Handle one command from the web server."""

        if len(read_bytes) != self.num_shafts // 8:
            self.log.warning(
                f"MockLoom: ignoring command {read_bytes!r}; length != {self.num_shafts // 8}"
            )
            return

        if not self.pick_wanted:  # type: ignore
            self.log.warning(
                f"MockLoom: ignoring command: {read_bytes!r}; pick not wanted"
            )
            return

        self.shaft_word = shaft_word_from_bytes(
            read_bytes, num_bytes=self.num_shafts // 8
        )
        self.pick_wanted = False
        await self.report_pick_wanted()
        await self.report_shafts()

    async def move(self, shaft_word: int) -> None:
        """This loom does not report shafts moving, so skip the usual code."""
        self.moving = False  # paranoia; it should never be set
        self.shaft_word = shaft_word
        await self.report_shafts()

    async def report_direction(self) -> None:
        if self.verbose:
            self.log.info(f"{self}.direction_forward={self.direction_forward}")

    async def report_motion_state(self) -> None:
        if self.verbose:
            self.log.info(f"{self}.moving={self.moving}")

    async def report_pick_wanted(self) -> None:
        if self.pick_wanted:
            reply = b"\x01" if self.direction_forward else b"\x02"
            await self.write(reply)

    async def report_shafts(self) -> None:
        if self.verbose:
            self.log.info(f"{self}.shaft_word={self.shaft_word}")



__all__ = ["LoomServer"]

import pathlib

from base_loom_server.base_loom_server import BaseLoomServer
from base_loom_server.enums import (
    DirectionControlEnum,
    MessageSeverityEnum,
    ShaftStateEnum,
)

from toika_loom_server.utils import bytes_from_shaft_word

from .mock_loom import MockLoom


class LoomServer(BaseLoomServer):
    """Communicate with the client software and the loom.

    The preferred way to create and run a LoomServer is to call
    LoomServer.amain(...).

    Args:
        num_shafts: The number of shafts the loom has.
        serial_port: The name of the serial port, e.g. "/dev/tty0".
            If the name is "mock" then use a mock loom.
        translation_dict: Language translation dict.
        reset_db: If True, delete the old database and create a new one.
        verbose: If True, log diagnostic information.
        db_path: Path to the pattern database. Specify None for the
            default path. Unit tests specify a non-None value, to avoid
            stomping on the real database.
    """

    default_name = "toika"
    loom_reports_motion = False
    loom_reports_direction = False
    mock_loom_type = MockLoom
    supports_full_direction_control = False

    def __init__(
        self,
        num_shafts: int,
        serial_port: str,
        translation_dict: dict[str, str],
        reset_db: bool,
        verbose: bool,
        db_path: pathlib.Path | None = None,
    ) -> None:
        super().__init__(
            num_shafts=num_shafts,
            serial_port=serial_port,
            translation_dict=translation_dict,
            reset_db=reset_db,
            verbose=verbose,
            db_path=db_path,
        )
        if self.loom_info.num_shafts % 8 != 0:
            raise ValueError(
                f"num_shafts={self.loom_info.num_shafts} must be a multiple of 8"
            )
        # The loom does not report motion state
        # so the shaft state is always DONE
        self.shaft_state = ShaftStateEnum.DONE

    async def basic_read_loom(self) -> bytes:
        """Read one reply_bytes from the loom.

        Perform no error checking, except that self.loom_reader exists.
        """
        assert self.loom_reader is not None
        return await self.loom_reader.readexactly(1)

    async def write_shafts_to_loom(self, shaft_word: int) -> None:
        """Send a shaft_word to the loom"""
        shaft_bytes = bytes_from_shaft_word(
            shaft_word=shaft_word, num_bytes=self.loom_info.num_shafts // 8
        )
        await self.write_to_loom(shaft_bytes)
        self.shaft_word = shaft_word

    async def handle_loom_reply(self, reply_bytes: bytes) -> None:
        """Process one reply_bytes from the loom."""
        # The only possible replies from the loom are:
        # b"\x01": request next forward pick
        # b"\x02": request next backward pick
        # TODO: allow the user to choose whether to use the loom's
        # commanded direction (respect the "reverse" button)
        # or the software's commanded direction.
        # For now only the software works.
        if reply_bytes not in {b"\x01", b"\x02"}:
            message = f"unexpected loom reply_bytes {reply_bytes!r}: should be b'\x01' or b'\x02'"
            self.log.warning(f"LoomServer: {message}")
            await self.report_command_problem(
                message=message,
                severity=MessageSeverityEnum.WARNING,
            )
            return

        if self.settings.direction_control == DirectionControlEnum.LOOM:
            # Loom controls weave direction, and we don't know the state
            # of the loom's direction button until it requests a new pick
            new_direction_forward = reply_bytes == b"\x01"
            if new_direction_forward != self.direction_forward:  # type: ignore
                self.direction_forward = new_direction_forward
                await self.report_direction()

        did_advance = await self.handle_next_pick_request()
        if did_advance:
            await self.report_shaft_state()



import importlib.resources

from base_loom_server.app_runner import AppRunner
from fastapi import FastAPI

from .loom_server import LoomServer

PKG_NAME = "toika_loom_server"
PKG_FILES = importlib.resources.files(PKG_NAME)

app = FastAPI()


app_runner = AppRunner(
    app=app,
    server_class=LoomServer,
    favicon=PKG_FILES.joinpath("favicon-32x32.png").read_bytes(),
    app_package_name=f"{PKG_NAME}.main:app",
)


def run_toika_loom() -> None:
    app_runner.run()



import asyncio
import contextlib
from collections.abc import AsyncGenerator

import pytest
from base_loom_server.base_mock_loom import BaseMockLoom
from base_loom_server.mock_streams import StreamReaderType, StreamWriterType

from toika_loom_server.mock_loom import MockLoom
from toika_loom_server.utils import bytes_from_shaft_word

# No need to reduce MockLoom.motion_duration to speed up tests
# because the Toika loom does not report motion state.


@contextlib.asynccontextmanager
async def create_loom(
    num_shafts: int = 16,
) -> AsyncGenerator[tuple[BaseMockLoom, StreamReaderType, StreamWriterType]]:
    """Create a MockLoom."""
    async with MockLoom(num_shafts=num_shafts, verbose=True) as loom:
        reader, writer = await loom.open_client_connection()
        yield loom, reader, writer


async def read(reader: StreamReaderType, timeout: float = 1) -> bytes:
    async with asyncio.timeout(timeout):
        return await reader.readexactly(1)


async def write(writer: StreamWriterType, command: bytes, timeout: float = 1) -> None:
    writer.write(command + MockLoom.terminator)
    async with asyncio.timeout(timeout):
        await writer.drain()


async def test_raise_shafts() -> None:
    async with create_loom(num_shafts=32) as (loom, reader, writer):
        for shaftword in (0x0, 0x1, 0x5, 0xFE, 0xFF19, 0xFFFFFFFE, 0xFFFFFFFF):
            # Tell mock loom to request the next pick
            await loom.oob_command("n")
            reply = await read(reader)
            assert reply == b"\x01"
            # Send the requested shaft information
            loom.command_event.clear()
            await write(
                writer, bytes_from_shaft_word(shaftword, num_bytes=loom.num_shafts // 8)
            )
            await loom.command_event.wait()
            assert loom.shaft_word == shaftword
        assert not loom.done_task.done()


async def test_oob_next_pick_and_toggle_direction() -> None:
    async with create_loom() as (loom, reader, writer):
        for expected_direction in (2, 1, 2, 1, 2, 1):
            await loom.oob_command("d")
            for _ in range(3):
                await loom.oob_command("n")
                reply = await read(reader)
                assert reply == expected_direction.to_bytes(length=1)
        assert not loom.done_task.done()


async def test_oob_close_connection() -> None:
    async with create_loom() as (loom, reader, writer):
        await loom.oob_command("c")
        async with asyncio.timeout(1):
            await loom.done_task
        assert loom.writer is not None
        assert loom.writer.is_closing()
        assert loom.reader is not None
        assert loom.reader.at_eof()


async def test_invalid_num_shafts() -> None:
    for bad_num_shafts in (0, 1, 7, 9, 15):
        with pytest.raises(ValueError):
            async with create_loom(num_shafts=bad_num_shafts) as (loom, reader, writer):
                pass



def test_version() -> None:
    try:
        import toika_loom_server.version
    except ImportError:
        raise AssertionError("version file not found")

    assert toika_loom_server.version.__all__ == ["__version__"]
    assert isinstance(toika_loom_server.version.__version__, str)
